using namespace std;
#include <config.h>
#include <ChiliLog.h>
#include <ChiliAudioChannel.h>
#include <ChiliProject.h>
#include <ChiliSession.h>
#include <ChiliFileDecoder.h>
#include <numtostr.h>
#include <FL/fl_ask.H>
#include <string>
#include <algorithm>
ChiliProject::ChiliProject(ChiliSession * p_session,const unsigned int p_slotnum, string proj_filename) {
	/* 
		Constructor called with proj_name implies that we 
		need to load the project from file.

		The filename can either be project file or simple file.

		If it is a nonexisting file, the project object
		can ask "nonexisting, create?" and create depending
		on the users needs.
	*/
	log("ChiliProject::ChiliProject");
	if (p_session==NULL) {
		lasterror=ERROR_SESSION_OBLIGATORY;
		return;	
	}
	log("ChiliProject::Session available.");
	session=p_session;
	
	slotnum=p_slotnum;
	log("Chiliproject::Creating project in slot %s",numtostr(slotnum).c_str());
	numberofchannels=0;
	workingfile=*new string("");
	filekey=*new string("");
	unsigned int ch;
	
	if (p_slotnum>=MAXPROJECTS) {
		lasterror=ERROR_INVALID_SLOT;		
	}
	
	/* Initialize channel list */
	for (ch=0;ch<MAXCHANNELS;ch++) {
		channellist[ch]=NULL;
	}
	log("ChiliProject::Channel list initialized.");	
	if (proj_filename=="") {
		/* NULL project filename, assume for now this is equal
		 * to an empty string. */
		return;
	}
	
	log("ChiliProject::Checking if file is simple or metafile");
	// project filename is given.
	if (isContainer(proj_filename)) {
		log("Item to load for project is a metafile.");
		LoadContainerFiles(proj_filename);
		return;
	}
	log("Item to load for project is a normal file (not a metafile).");

	// Filename is given and it is not a container.
	// Attempt to load the file.
	ChiliFileDecoder * simplefile;
	lasterror=0;
	try {
		simplefile=new ChiliFileDecoder(session); 	/* File decoder object */
	}
	catch (int e) {
		throw ERROR_CREATING_FILEOBJECT;
	}

	log("About to load sourcefile");
	try {
		/* Loading the source file means decoding it to a raw format.
		 * At some point we want to have peak info for this file;
		 * the most efficient way is to do the peak scanning during
		 * decoding of the file, so this is done here as well.
		 */
		simplefile->LoadSourceFile(proj_filename);
	}
	catch (int e) {
		throw e;
	}

	// TODO: All is ok, all file decoding done.
	
	unsigned int chnum=0;
	unsigned int addchan=0;

	unsigned int channels=0;
	channels=simplefile->GetNumberOfChannels();
	log("ChiliProject::found %s channels",numtostr(channels).c_str());

	workingfile=simplefile->GetDecodedFilename();
	if (workingfile=="") {
		log("ChiliProject::no working file");
		lasterror=ERROR_FILENAME_OBLIGATORY;
		return;
	}
	
	filekey=simplefile->GetDecodedFilekey();
	if (filekey=="") {
		log("ChiliProject::no working file key");
		lasterror=ERROR_FILENAME_OBLIGATORY;
		return;
	}

	Fl_Preferences * userprefs=session->GetUserPrefs();
	
	string project="Project"+numtostr(slotnum);
	string projdotsource=project+".sourcefile";
	string projdotkey=project+".filekey";
	string projdotworkfile=project+".workingfile";

	userprefs->set(projdotsource.c_str(),proj_filename.c_str());
	userprefs->set(projdotkey.c_str(),filekey.c_str());  // relate file key to full filename
	userprefs->set(projdotworkfile.c_str(),workingfile.c_str());  // relate file key to full filename
	userprefs->flush();

	for (chnum=0;chnum<channels;chnum++) {
		// Add each decoded file to the project as audio channel.

		addchan=AddChannel();
		if (addchan==MAXCHANNELS) {
			log("ChiliProject::could not add channels");
			// Could not add channel. Don't modify lasterror, 
			// propagate whatever caused this.
			return;
		}
	}
	/* 
	   All loading and decoding is done, all raw pcm files are assigned
	   to a channel. This is where the responsibility of the file decoder
	   object ends. Further handling and cleaning up of temp files is the
           responsibility of each individual AudioChannel object.
	*/
	delete simplefile;
	return;
}

ChiliProject::~ChiliProject() {
	/* Clean up the project. First destroy any channels
	 * that the project contains. */
	unsigned int i;
	ChiliAudioChannel * ca;
	log("ChiliProject::~ChiliProject");
	for (i=0;i<MAXCHANNELS;i++) {
		if (channellist[i]==NULL) continue;
		ca=channellist[i];
		delete ca;
		channellist[i]=NULL;
	}

	if (slotnum>=MAXPROJECTS) {
		lasterror=ERROR_INVALID_SLOT;
		return;
	}
	if (session==NULL) {
		lasterror=ERROR_SESSION_OBLIGATORY;
		return;
	}

	session->projectlist[slotnum]=NULL;

	string project="Project"+numtostr(slotnum);
	string projdotsource=project+".sourcefile";
	string projdotkey=project+".filekey";
	string projdotworkfile=project+".workingfile";

	unlink(workingfile.c_str());

	Fl_Preferences * userprefs=session->GetUserPrefs();

	if (userprefs->entryExists(projdotsource.c_str())) {
		userprefs->deleteEntry(projdotsource.c_str());
	}
	if (userprefs->entryExists(projdotkey.c_str())) {
		userprefs->deleteEntry(projdotkey.c_str());
	}
	if (userprefs->entryExists(projdotworkfile.c_str())) {
		userprefs->deleteEntry(projdotworkfile.c_str());
	}
	userprefs->flush();
	return;
}

unsigned int ChiliProject::GetNumberOfChannels() {
	return numberofchannels;
}

bool ChiliProject::isContainer(string proj_filename) {
	log("ChiliProject::isContainer");
	// TODO: Implement function for project file load support.
	// Until then this function will return false.
	// This will result in the simple file loader being
	// invoked, who in turn will tell the user that
	// the file format is not supported.
	if (proj_filename=="") {
		return false;
	}
	
	return false;	
}
void ChiliProject::LoadContainerFiles(string proj_filename) {
	/* Likewise the implementation of this function is
	   not yet needed. */
	log("ChiliProject::LoadContainerFiles");
	fl_message("loadcontainerfiles(not implemented yet)");
       	throw ERROR_IMPLEMENTATION_INCOMPLETE;
	return;
}

unsigned int ChiliProject::SetSampleRate(const unsigned int rate) {
	log("ChiliProject::SetSampleRate");
	samplerate=rate;
	return samplerate;
}

unsigned int ChiliProject::GetFreeChannel() {
	log("ChiliProject::GetFreeChannel");
	unsigned int channelnum=0;
	while (channelnum<MAXCHANNELS) {
		if (channellist[channelnum]==NULL) return channelnum;
		channelnum++;
	}
	log("ChiliProject::GetFreeChannel(): No free channels");
	return channelnum;
}

unsigned int ChiliProject::AddChannel() {
	/* Attempt to add a new (empty) channel to the project.
	   Returns zero based number of created channel,
	   or MAXCHANNELS otherwise. */
	ChiliAudioChannel * channel=NULL;
	log("ChiliProject::AddChannel");

	unsigned int freechannel=GetFreeChannel();
	if (freechannel>=MAXCHANNELS) {
		lasterror=ERROR_NO_FREE_CHANNELS;
		log("ChiliProject::AddChannel() : No free channels");
		return MAXCHANNELS;
	}
	channel=new ChiliAudioChannel(this,freechannel);

	if (channel==NULL) {
		/* channel object creation failed. */
		log("ChiliProject::AddChannel() : Channel object creation failed.");
		if (lasterror==0) {
			/* Unknown reason. */
			lasterror=ERROR_CREATE_CHANNEL_FAILED;
		}
		return MAXCHANNELS;
	}
	log ("ChiliProject::AddChannel() : All seems OK.");
	numberofchannels++;
	channellist[freechannel]=channel;
	return freechannel;
}

float ChiliProject::GetMinPeak(int channel,int pos,int maxpos,float zoom) {
	/* fs=filesize 		(e.g. 50 000 000)
	 * vss=virtual screensize=maxpos+1
	 * bs=blocksize	in real samples=fs/vss
	 * pr=peak ratio (for now we assume a peak ratio of 160, or
	 *                1 value in peak file corresponds to 160 values
	 *		  in the original file).
	 * pv=number of peak values to calculate
	 * 
	 * bs=fs/(maxpos+1)	(e.g. 50 000 000 / 2000 = 25000)
	 * pv=(bs/pr)
         */
	string strR=numtostr(channel)+",pos="+numtostr(pos)+",maxpos="+numtostr(maxpos)+",zoom="+numtostr(zoom);
//	log("ChiliProject::GetMinPeak(channel=%s)",strR.c_str());

	float temp=(2*rand()/(RAND_MAX+1.0));
	return temp;
}

float ChiliProject::GetMaxPeak(int channel,int pos,int maxpos,float zoom) { 
	string strR=numtostr(channel)+",pos="+numtostr(pos)+",maxpos="+numtostr(maxpos)+",zoom="+numtostr(zoom);
//	log("ChiliProject::GetMaxPeak(channel=%s)",strR.c_str());
	float temp=(2*rand()/(RAND_MAX+1.0));
	return temp;
}
